/*
    View64 C64 image viewer
    (c) 1995-2013 Soci/Singular (soci at c64.rulez.org)

    $Id: view64-sdl1.2.c 27 2014-04-06 20:52:33Z soci $

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License along
    with this program; if not, write to the Free Software Foundation, Inc.,
    51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

*/

#include <stdio.h>
#include <string.h>
#include <FL/Fl.H>
#include <FL/Fl_Window.H>
#include <FL/fl_draw.H>
#include <FL/Fl_Menu_Bar.H>
#include <FL/Fl_Scrollbar.H>
#include <FL/Fl_Box.H>
#include <FL/fl_ask.H>
#include <FL/fl_utf8.h>
#include <FL/Fl_Native_File_Chooser.H>

#include "libview64.h"
#include "tvrender.h"
#include "view64_config.h"
#include "main.h"
#include "cfgpath.h"
#include "ini.h"

static int NORMAL_SIZE, MY_SIZE;

class Main_window;

static Main_window *main_window;
static arguments_t arguments;

static void word(unsigned char *p, int w) {p[0] = w; p[1] = w >> 8;}
static void dword(unsigned char *p, int w) {word(p, w);word(p + 2, w >> 16);}

static void bmp_save(const char *name, const unsigned char *map, int w, int h) {
    int x;
    unsigned char header[54];
    FILE *f1;
    f1 = fl_fopen(name, "wb");
    if (f1 == NULL) {
        return;
    }

    word(header + 0, 0x4D42);
    dword(header + 2, ((w*3+3) & ~3)*h + sizeof(header));
    dword(header + 6, 0);
    dword(header + 10, sizeof(header));

    dword(header + 14, sizeof(header) - 14);
    dword(header + 18, w);
    dword(header + 22, h);
    word(header + 26, 1);
    word(header + 28, 24);
    dword(header + 30, 0);
    dword(header + 34, ((w*3+3) & ~3)*h);
    dword(header + 38, 0);
    dword(header + 42, 0);
    dword(header + 46, 0);
    dword(header + 50, 0);

    fwrite(header, 1, sizeof(header), f1);
    memset(header, 0, 3);
    while (h--) {
        const  unsigned char *p = map + w * 3 * h;
        for (x = 0; x < w; x++, p += 3) {
            putc(p[2], f1);
            putc(p[1], f1);
            putc(p[0], f1);
        }
        fwrite(header, 1, ((-w * 3)) & 3, f1);
    }
    fclose(f1);
}

static view64_screen *open_file(const char *filename, view64_format *format) {
    FILE *f1;
    size_t size;
    unsigned char *mem;
    view64_screen *rscreen;

    *format = VIEW64_UNKNOWN;
    if (filename == NULL || !filename[0]) {
        return NULL;
    }
    if ((f1 = fl_fopen(filename, "rb")) == NULL) {
        fl_alert("Couldn't open file");
        return NULL;
    }

    if (fseek(f1, 0, SEEK_END)) {
        size = 65536;
    } else {
        size = ftell(f1);
        if (size < 65536) size = 65536;
    }
    rewind(f1);
    mem = (unsigned char *)calloc(1, size);
    if (mem == NULL) {
        fclose(f1);
        return NULL;
    }

    size = fread(mem, 1, size, f1);
    fclose(f1);
    *format = view64_format_detect(filename, mem, size);
    if (*format == VIEW64_UNKNOWN) {
        fl_alert("Image format not recognized");
        free(mem);
        return NULL;
    }
    rscreen = view64_decode(*format, mem, size, 0);
    free(mem);
    if (rscreen == NULL) {
        fl_alert("Can't decode image");
        return NULL;
    }
    return rscreen;
}

class Main_window : public Fl_Window {
    class Drop : public Fl_Box {
        static void open(void *a) {
            char *txt = (char *)a, *name;
            if (!strncmp(txt, "file://", 7)) {
                fl_decode_uri(txt);
                name = txt + 7; 
            } else name = txt;
            free(arguments.filename);
            arguments.filename = (char *)malloc(strlen(name) + 1);
            strcpy(arguments.filename, name);
            main_window->reopen_file();
            main_window->redraw();
            main_window->update_menu();
            free(txt);
        }
        public:
        Drop(int x, int y, int w, int h) : Fl_Box(x, y, w, h) {
        }
        int handle(int e) {
            size_t len;
            switch (e) {
            case FL_PASTE:
                len = Fl::event_length();
                if (len) {
                    char *txt = (char *)malloc(len + 1);
                    if (txt) {
                        strcpy(txt, Fl::event_text());
                        if (len && txt[len-1] == 0x0a) len--;
                        if (len && txt[len-1] == 0x0d) len--;
                        txt[len] = 0;
                        Fl::add_timeout(0.0, open, txt);
                    }
                }
            case FL_DND_RELEASE: 
            case FL_DND_ENTER:
            case FL_DND_DRAG:
            case FL_DND_LEAVE: 
                return 1;
            default: return Fl_Box::handle(e);
            }
        }
    };

    class Picture : public Fl_Group {
        int texture_scale, new_palette, new_picture;
        unsigned int texture_video;
        rgb_color *palette;
        Drop *dummy;
        Fl_Scrollbar *vscroll, *hscroll;
        int push_x, push_y;
        public:
        unsigned char *surface;
        int pw, ph;
        view64_screen *rscreen;
        int lastzoom;

        Picture(int x, int y, int w, int h, view64_screen *pic, rgb_color *pal) : Fl_Group(x, y, w, h) {
            Fl_Group::box(FL_NO_BOX);
            rscreen = pic;
            surface = NULL;
            texture_scale = 0;
            texture_video = VIDEO_RGB;
            new_palette = 1;
            new_picture = 1;
            palette = pal;
            dummy = new Drop(x, y, w - FL_NORMAL_SIZE, h - FL_NORMAL_SIZE);
            resizable(dummy);
            vscroll = new Fl_Scrollbar(x + w - FL_NORMAL_SIZE, y, FL_NORMAL_SIZE, h - FL_NORMAL_SIZE);
            vscroll->bounds(-32768, 32767);
            vscroll->linesize(4096);
            vscroll->type(FL_VERTICAL);
            vscroll->callback(scroll_cb, this);
            hscroll = new Fl_Scrollbar(x, y + h - FL_NORMAL_SIZE, w - FL_NORMAL_SIZE, FL_NORMAL_SIZE);
            hscroll->bounds(-32768, 32767);
            hscroll->linesize(4096);
            hscroll->type(FL_HORIZONTAL);
            hscroll->callback(scroll_cb, this);
            lastzoom = arguments.zoom;
        }

        ~Picture() {
            delete dummy;
            delete vscroll;
            delete hscroll;
            free(surface);
            if (rscreen) {
                view64_free_screen(rscreen);
            }
        }

        void set_palette(rgb_color *pal) {
            palette = pal;
            new_palette = 1;
        }

        void set_picture(view64_screen *scr) {
            if (rscreen) {
                view64_free_screen(rscreen);
            }
            rscreen = scr;
            new_picture = 1;
        }
        private:
        static void scroll_cb(Fl_Widget *w, void *a) {
            ((Picture *)a)->redraw();
        }

        int handle(int e) {
            switch (e) {
            case FL_PUSH:
                if (Fl::event_state() != FL_BUTTON1) break;
                push_x = Fl::event_x();
                push_y = Fl::event_y();
                if (hscroll->visible() || vscroll->visible()) {
                    main_window->cursor(FL_CURSOR_MOVE);
                    return 1;
                }
                break;
            case FL_DRAG:
                if (hscroll->visible() && pw > hscroll->w()) {
                    double w = hscroll->value() + (push_x - Fl::event_x()) * 65536.0 / (pw - hscroll->w());
                    if (w < -32768) w = -32768;
                    if (w > 32767) w = 32767;
                    hscroll->value(w);
                    push_x = Fl::event_x();
                }
                if (vscroll->visible() && ph > hscroll->h()) {
                    double h = vscroll->value() + (push_y - Fl::event_y()) * 65536.0 / (ph - vscroll->h());
                    if (h < -32768) h = -32768;
                    if (h > 32767) h = 32767;
                    vscroll->value(h);
                    push_y = Fl::event_y();
                }
                redraw();
                return 1;
            case FL_RELEASE:
                main_window->cursor(FL_CURSOR_DEFAULT);
                return 1;
            case FL_MOUSEWHEEL:
                switch (Fl::event_state()) {
                case FL_SHIFT:
                    if (Fl::event_dy() && hscroll->visible()) {
                        double w = hscroll->value();
                        w += hscroll->linesize() * Fl::event_dy();
                        if (w > 32767) w = 32767;
                        if (w < -32767) w = -32767;
                        hscroll->value(w);
                        redraw();
                        return 1;
                    }
                    break;
                default: break;
                }
                break;
            }
            return Fl_Group::handle(e);
        }

        void draw() {
            int scale, scale2, bx, by;
            int w = Fl_Group::w(), h = Fl_Group::h();
            int x = Fl_Group::x(), y = Fl_Group::y();
            int w2, h2;

            if (rscreen == NULL) {
                fl_color(FL_BACKGROUND_COLOR);
                fl_rectf(x, y, w, h);
                hscroll->hide();
                vscroll->hide();
                return;
            }

            if (arguments.zoom) {
                scale = arguments.zoom;
            } else {
                scale = w / rscreen->w;
                scale2 = h / rscreen->h;
                if (scale2 < scale) scale = scale2;
                if (rscreen->w * scale > w || rscreen->h * scale > h) {
                    scale--;
                }
                if (scale < 1) scale = 1;
            }

            ph = rscreen->h * scale;
            pw = rscreen->w * scale;
            if (pw > w) {
                h2 = h - FL_NORMAL_SIZE;
                hscroll->show();
                if (ph > h2) {
                    vscroll->show();
                    vscroll->size(FL_NORMAL_SIZE, h - FL_NORMAL_SIZE);
                    hscroll->size(w - FL_NORMAL_SIZE, FL_NORMAL_SIZE);
                    w2 = w - FL_NORMAL_SIZE;
                } else {
                    vscroll->hide();
                    hscroll->size(w, FL_NORMAL_SIZE);
                    w2 = w;
                }
            } else if (ph > h) {
                w2 = w - FL_NORMAL_SIZE;
                vscroll->show();
                if (pw > w2) {
                    hscroll->show();
                    vscroll->size(FL_NORMAL_SIZE, h - FL_NORMAL_SIZE);
                    hscroll->size(w - FL_NORMAL_SIZE, FL_NORMAL_SIZE);
                    h2 = h - FL_NORMAL_SIZE;
                } else {
                    hscroll->hide();
                    vscroll->size(FL_NORMAL_SIZE, h);
                    h2 = h;
                }
            } else {
                hscroll->hide();
                vscroll->hide();
                w2 = w; h2 = h;
            }
            bx = (w2 - pw) / 2;
            by = (h2 - ph) / 2;

            if (hscroll->visible() || vscroll->visible()) {
                if (hscroll->visible()) {
                    hscroll->slider_size((float)w2 / pw);
                    bx -= hscroll->value() * (pw - w2) / 65536;
                }
                if (vscroll->visible()) {
                    vscroll->slider_size((float)h2 / ph);
                    by -= vscroll->value() * (ph - h2) / 65536;
                }
                Fl_Group::draw();
                if (hscroll->visible() && vscroll->visible()) {
                    fl_color(FL_BACKGROUND_COLOR);
                    fl_rectf(x + w - FL_NORMAL_SIZE, y + h - FL_NORMAL_SIZE, FL_NORMAL_SIZE, FL_NORMAL_SIZE);
                }
            }
            fl_push_clip(x, y, w - (vscroll->visible() ? FL_NORMAL_SIZE : 0), h - (hscroll->visible() ? FL_NORMAL_SIZE : 0));

            if (texture_scale != scale || texture_video != arguments.vmode || new_palette || new_picture) {
                tvrender_image out;
                out.h = ph;
                out.w = pw;
                out.pitch = out.w * 3;
                surface = (unsigned char*)realloc(surface, out.h * out.pitch);
                if (surface) {
                    out.d = surface;
                    create_tvpicture(rscreen, arguments.vmode, scale, palette, &out);
                }
                texture_scale = scale;
                texture_video = arguments.vmode;
                new_palette = 0;
                lastzoom = scale;
            }
            fl_color(FL_BLACK);
            fl_rectf(x, y, w, by);
            fl_rectf(x, y + by, bx, ph);
            fl_rectf(x + bx + pw, y + by, w - bx - pw, ph);
            fl_rectf(x, y + by + ph, w, h - by - ph);
            fl_draw_image(surface, x + bx, y + by, pw, ph, 3, 0);
            fl_pop_clip();
        }
    };
    char *window_title;
    Picture *picture;
    Fl_Menu_Bar *menu;
    rgb_color *palette;
    int scale, scale2;
    struct {
        int x, y, w, h;
    } windim;

public:
    Main_window() : Fl_Window(0, 0, 320, 200 + (arguments.menu ? FL_NORMAL_SIZE * 2 : 0), "View64 " VERSION) {
        int x, y, w, h;
        FILE *f1;
        view64_screen *scr;
        static Fl_Menu_Item menu_[] = {
            {"&File", 0,  0, 0, FL_SUBMENU},
            {"&Open ", 'o', open_cb},
            {"&Save ", 's', save_cb, NULL},
            {"&Quit ", FL_Escape, quit_cb, NULL},
            {0},
            {"&Options", 0,  0, 0, FL_SUBMENU},
            {"&Decoder", 0,  0, 0, FL_SUBMENU | FL_MENU_DIVIDER},
            {0},
            {0},
            {"&Zoom", 0,  0, 0, FL_SUBMENU},
            {"Zoom &in ", '+', zoom2_cb, (void *)3},
            {"Zoom &out ", '-', zoom2_cb, (void *)1},
            {0},
            {"&Window", 0,  0, 0, FL_SUBMENU},
            {"&Fullscreen "},
            {"&Menu bar "},
            {"Re&size ", 'z', resize_cb},
            {"&Remember", 0, winsave_cb, (void *)1},
            {"&Forget"},
            {"&About", 0, about_cb},
            {0},
        };

        window_title = NULL;
        picture = NULL;
        f1 = arguments.palette_filename && arguments.palette_filename[0] ? fl_fopen(arguments.palette_filename, "rb") : NULL;
        palette = load_palette(f1);
        if (f1) fclose(f1);
        scr = reopen_file();

        w = arguments.width > 4160 ? 4160 : arguments.width;
        h = arguments.height > 2600 ? 2600 : arguments.height;
        x = arguments.x < 0 ? Fl::w() / 2 : arguments.x; 
        y = arguments.y < 0 ? Fl::h() / 2 : arguments.y; 

        if (w < 1) {
            if (h < 1) {
                int scale = arguments.zoom;
                w = scr ? scr->w : 320;
                h = scr ? scr->h : 200;
                if (!scale) {
                    scale = 1;
                    int sw = Fl::w() * 0.618, sh = Fl::h() * 0.618 - FL_NORMAL_SIZE * 2;
                    while (w * (scale + 1) < sw && h * (scale + 1) < sh && scale < 13) {
                        scale++;
                    }
                }
                w *= scale;
                h *= scale;
            } else {
                if (scr) {
                    w = scr->w * h / scr->h;
                } else {
                    w = 640 * h / 400;
                }
            }
        } else if (h < 1) {
            if (scr) {
                h = scr->h * w / scr->w;
            } else {
                h = 400 * w / 640;
            }
        } else {
            /* all set */
        }

        begin();
        Fl_Window::box(FL_NO_BOX);
        menu = new Fl_Menu_Bar(0, 0, 320, arguments.menu ? FL_NORMAL_SIZE * 2 : 0);
        menu->menu(menu_);
        picture = new Picture(0, arguments.menu ? FL_NORMAL_SIZE * 2 : 0, 320, 200, scr, palette);
        resizable(picture);
        end();
        resize(x - w / 2, y - h / 2 - (arguments.menu ? FL_NORMAL_SIZE : 0), w, h + (arguments.menu ? FL_NORMAL_SIZE * 2 : 0));
        if (arguments.fullscreen) {
            fullscreen();
        }
        if (arguments.filename == NULL || arguments.filename[0] == 0 || scr != NULL) {
            show();
        }
        update_menu();
    }

    ~Main_window() {
        delete picture;
        free(palette);
        free(window_title);
    }

    void hide_menu(void) {
        if (menu->h()) {
            menu->size(menu->w(), 0);
            picture->resize(0, 0, picture->w(), picture->h() + FL_NORMAL_SIZE * 2);
            init_sizes();
        }
    }

    void show_menu(void) {
        if (!menu->h() && arguments.menu) {
            picture->resize(0, FL_NORMAL_SIZE * 2, picture->w(), picture->h() - FL_NORMAL_SIZE * 2);
            menu->size(menu->w(), FL_NORMAL_SIZE * 2);
            init_sizes();
        }
    }

    void fullscreen(void) {
        hide_menu();
        Fl_Window::fullscreen();
    }
    void fullscreen_off(void) {
        Fl_Window::fullscreen_off();
        show_menu();
    }
private:
    void update_menu(void) {
        menu->add("File/Save ", 's', save_cb, NULL, (arguments.filename && arguments.filename[0] && picture && picture->rscreen ? 0 : FL_MENU_INACTIVE));
        menu->add("Options/&Custom palette", 0, palette_cb, (void *)0, ((arguments.vmode & VIDEO_MASK) != VIDEO_RGB ? FL_MENU_INACTIVE : 0) | FL_MENU_RADIO | (arguments.palette_filename && arguments.palette_filename[0] ? FL_MENU_VALUE : 0));
        menu->add("Options/&Default palette", 0, palette_cb, (void *)1, ((arguments.vmode & VIDEO_MASK) != VIDEO_RGB ? FL_MENU_INACTIVE : 0) | FL_MENU_RADIO | FL_MENU_DIVIDER| (arguments.palette_filename && arguments.palette_filename[0] ? 0 : FL_MENU_VALUE));
        menu->add("Options/&Chroma noise ", 'c', effect_cb, (void *)VIDEO_CHROMA_NOISE, FL_MENU_TOGGLE | ((arguments.vmode & VIDEO_CHROMA_NOISE) ? FL_MENU_VALUE : 0) | ((arguments.vmode & VIDEO_MASK) == VIDEO_RGB ? FL_MENU_INACTIVE : 0));
        menu->add("Options/MC&I deinterlace ", 'i', effect_cb, (void *)VIDEO_NOLACE, FL_MENU_TOGGLE | ((arguments.vmode & VIDEO_NOLACE) ? FL_MENU_VALUE : 0) | ((arguments.vmode & VIDEO_MASK) != VIDEO_RGB ? FL_MENU_INACTIVE : 0));
        menu->add("Options/Scan&lines ", 'l', effect_cb, (void *)VIDEO_SCANLINES, FL_MENU_TOGGLE | ((arguments.vmode & VIDEO_SCANLINES) ? FL_MENU_VALUE : 0));
        menu->add("Options/&TV mask ", 't', effect_cb, (void *)VIDEO_SHADOWMASK, FL_MENU_TOGGLE | ((arguments.vmode & VIDEO_SHADOWMASK) ? FL_MENU_VALUE : 0));
        menu->add("Options/&Monochrome ", 'm', effect_cb, (void *)VIDEO_MONOCHROME, FL_MENU_TOGGLE | ((arguments.vmode & VIDEO_MONOCHROME) ? FL_MENU_VALUE : 0));
        menu->add("Options/Decoder/&RGB ", 'r', decoder_cb, (void *)VIDEO_RGB, FL_MENU_RADIO | ((arguments.vmode & VIDEO_MASK) == VIDEO_RGB ? FL_MENU_VALUE : 0));
        menu->add("Options/Decoder/&PAL ", 'p', decoder_cb, (void *)VIDEO_PAL, FL_MENU_RADIO | ((arguments.vmode & VIDEO_MASK) == VIDEO_PAL ? FL_MENU_VALUE : 0));
        menu->add("Options/Decoder/&NTSC ", 'n', decoder_cb, (void *)VIDEO_NTSC, FL_MENU_RADIO | ((arguments.vmode & VIDEO_MASK) == VIDEO_NTSC ? FL_MENU_VALUE : 0));
        menu->add("Zoom/Auto ", '0', zoom_cb, (void *)0, FL_MENU_RADIO | (arguments.zoom  == 0 ? FL_MENU_VALUE : 0));
        menu->add("Zoom/&1x ", '1', zoom_cb, (void *)1, FL_MENU_RADIO | (arguments.zoom  == 1 ? FL_MENU_VALUE : 0));
        menu->add("Zoom/&2x ", '2', zoom_cb, (void *)2, FL_MENU_RADIO | (arguments.zoom  == 2 ? FL_MENU_VALUE : 0));
        menu->add("Zoom/&3x ", '3', zoom_cb, (void *)3, FL_MENU_RADIO | (arguments.zoom  == 3 ? FL_MENU_VALUE : 0));
        menu->add("Zoom/&4x ", '4', zoom_cb, (void *)4, FL_MENU_RADIO | (arguments.zoom  == 4 ? FL_MENU_VALUE : 0));
        menu->add("Zoom/&5x ", '5', zoom_cb, (void *)5, FL_MENU_RADIO | (arguments.zoom  == 5 ? FL_MENU_VALUE : 0));
        menu->add("Zoom/&6x ", '6', zoom_cb, (void *)6, FL_MENU_RADIO | (arguments.zoom  == 6 ? FL_MENU_VALUE : 0));
        menu->add("Zoom/&7x ", '7', zoom_cb, (void *)7, FL_MENU_RADIO | (arguments.zoom  == 7 ? FL_MENU_VALUE : 0));
        menu->add("Zoom/&8x ", '8', zoom_cb, (void *)8, FL_MENU_RADIO | (arguments.zoom  == 8 ? FL_MENU_VALUE : 0));
        menu->add("Zoom/&9x ", '9', zoom_cb, (void *)9, FL_MENU_RADIO | (arguments.zoom  == 9 ? FL_MENU_VALUE : 0));
        menu->add("Window/Forget", 0, winsave_cb, (void *)0, (arguments.x < 0 && arguments.y < 0 && arguments.width < 1 && arguments.height < 1 ? FL_MENU_INACTIVE : 0));
        menu->add("Window/Fullscreen ", 'f', fullscreen_cb, NULL, FL_MENU_TOGGLE | (arguments.fullscreen ? FL_MENU_VALUE : 0));
        menu->add("Window/Menu bar ", FL_COMMAND + 'm', menubar_cb, NULL, FL_MENU_TOGGLE | (arguments.menu ? FL_MENU_VALUE : 0));
    }

    void setzoom(int zoom) {
        arguments.zoom = zoom;
        picture->redraw();
        update_menu();
    }

    static void zoom_cb(Fl_Widget *w, void *a) {
        main_window->setzoom((long)a);
    }

    static void zoom2_cb(Fl_Widget *w, void *a) {
        int i = main_window->picture->lastzoom + (long)a - 2;
        if (i < 1) i = 1;
        if (i > 9) i = 9;
        main_window->setzoom(i);
    }

    void toggle_effect(unsigned int effect) {
        arguments.vmode ^= effect;
        picture->redraw();
        update_menu();
    }

    static void effect_cb(Fl_Widget *w, void *a) {
        main_window->toggle_effect((unsigned long)a);
    }

    void set_decoder(unsigned int video) {
        arguments.vmode = (arguments.vmode & ~VIDEO_MASK) | video;
        picture->redraw();
        update_menu();
    }

    static void decoder_cb(Fl_Widget *w, void *a) {
        main_window->set_decoder((unsigned long)a);
    }

    void toggle_fullscreen(void) {
        arguments.fullscreen ^= 1;
        if (arguments.fullscreen) {
            fullscreen();
        } else {
            fullscreen_off();
        }
    }

    static void fullscreen_cb(Fl_Widget *w, void *a) {
        main_window->toggle_fullscreen();
    }

    static void winsave_cb(Fl_Widget *w, void *a) {
        if (a) {
            arguments.width = main_window->picture->w();
            arguments.height = main_window->picture->h();
            arguments.x = main_window->x() + main_window->w() / 2;
            arguments.y = main_window->y() + main_window->h() / 2;
        } else {
            arguments.width = 0;
            arguments.height = 0;
            arguments.x = -1;
            arguments.y = -1;
            main_window->update_menu();
        }
    }

    static void palette_cb(Fl_Widget *w, void *a) {
        const char *name;
        if (a) {
            free(arguments.palette_filename);
            arguments.palette_filename = NULL;
        } else {
            FL_NORMAL_SIZE = NORMAL_SIZE;
            int ret;
            Fl_Native_File_Chooser f;
            f.title("Select palette file");
            f.type(Fl_Native_File_Chooser::BROWSE_FILE);
            f.filter("VICE palette\t*.vpl");
            if (arguments.palette_filename && arguments.palette_filename[0]) {
                f.preset_file(arguments.palette_filename);
            }
            ret = f.show();
            FL_NORMAL_SIZE = MY_SIZE;
            switch (ret) {
            case -1: 
            case  1:
                break;
            default: 
                free(arguments.palette_filename);
                name = f.filename();
                arguments.palette_filename = (char *)malloc(strlen(name) + 1);
                strcpy(arguments.palette_filename, name);
            }
        }
        main_window->set_palette(arguments.palette_filename);
    }

    void set_palette(const char *name) {
        FILE *f1;
        free(palette);
        f1 = arguments.palette_filename && arguments.palette_filename[0] ? fl_fopen(arguments.palette_filename, "rb") : NULL;
        palette = load_palette(f1);
        if (f1) fclose(f1);
        picture->set_palette(palette);
        picture->redraw();
        update_menu();
    }

    static void save_cb(Fl_Widget *w, void *a) {
        if (arguments.filename && arguments.filename[0] && main_window->picture->surface) {
            size_t len = strlen(arguments.filename);
            char *newname = (char *)malloc(len + 5);
            int ret;
            if (newname == NULL) return;
            FL_NORMAL_SIZE = NORMAL_SIZE;
            Fl_Native_File_Chooser f;
            strcpy(newname, arguments.filename);
            if (len >= 4 && newname[len-4] == '.') {
                len -= 4;
            }
            strcpy(newname + len, ".bmp");
            f.title("Save image file");
            f.type(Fl_Native_File_Chooser::BROWSE_SAVE_FILE);
            f.filter("Device Independent Bitmap\t*.bmp");
            f.preset_file(newname);
            ret = f.show();
            FL_NORMAL_SIZE = MY_SIZE;
            switch (ret) {
            case -1: 
            case  1:
                break;
            default: 
                bmp_save(f.filename(), main_window->picture->surface, main_window->picture->pw, main_window->picture->ph);
            }
            free(newname);
        }
    }

    static void quit_cb(Fl_Widget *w, void *a) {
        main_window->hide();
    }

    view64_screen *reopen_file(void) {
        view64_format format;
        view64_screen *scr = open_file(arguments.filename, &format);
        free(window_title);
        if (scr) {
            const char *format_name = view64_format_name(format);
            window_title = (char *)malloc(strlen(format_name) + strlen("View64 " VERSION " - ") + 4 + strlen(arguments.filename));
            if (window_title) {
                strcpy(window_title, "View64 " VERSION " - ");
                strcat(window_title, format_name);
                strcat(window_title, " - ");
                strcat(window_title, fl_filename_name(arguments.filename));
            }
        } else {
            window_title = NULL;
        }
        label(window_title ? window_title : "View64 " VERSION);
        if (picture) picture->set_picture(scr);
        return scr;
    }

    static void open_cb(Fl_Widget *w, void *a) {
        const char *name;
        int ret;
        FL_NORMAL_SIZE = NORMAL_SIZE;
        Fl_Native_File_Chooser f;
        f.title("Select picture file");
        f.type(Fl_Native_File_Chooser::BROWSE_FILE);
        ret = f.show();
        FL_NORMAL_SIZE = MY_SIZE;
        switch (ret) {
        case -1: 
        case  1:
            break;
        default: 
            free(arguments.filename);
            name = f.filename();
            arguments.filename = (char *)malloc(strlen(name) + 1);
            strcpy(arguments.filename, name);
            main_window->reopen_file();
            main_window->redraw();
            main_window->update_menu();
            break;
        }
    }

    static void about_cb(Fl_Widget *w, void *a) {
        fl_message("View64 " VERSION "\n" 
                "(C) 1995-2015 Soci/Singular (soci at c64.rulez.org)\n\n"
                "http://view64.sourceforge.net/\n\n"
                "This program is free software; you can redistribute it and/or modify\n"
                "it under the terms of the GNU General Public License as published by\n"
                "the Free Software Foundation; either version 2 of the License, or\n"
                "(at your option) any later version.\n\n"
                "So if you don't like the GUI, go and fix it ;)");
    }

    static void menubar_cb(Fl_Widget *w, void *a) {
        arguments.menu ^= 1;
        if (!main_window->fullscreen_active()) {
            if (arguments.menu) {
                main_window->show_menu();
                main_window->resize(main_window->x(), main_window->y() - FL_NORMAL_SIZE, main_window->w(), main_window->h() + FL_NORMAL_SIZE * 2);
            } else {
                main_window->hide_menu();
                main_window->resize(main_window->x(), main_window->y() + FL_NORMAL_SIZE, main_window->w(), main_window->h() - FL_NORMAL_SIZE * 2);
            }
            main_window->redraw();
        }
    }

    static void resize_cb(Fl_Widget *wg, void *a) {
        int x = main_window->x();
        int y = main_window->y();
        int w = main_window->w();
        int h = main_window->h();
        int w2;
        int h2;

        if (!main_window->fullscreen_active()) {
            if (arguments.zoom) {
                w2 = main_window->picture->rscreen ? main_window->picture->rscreen->w : 320;
                h2 = main_window->picture->rscreen ? main_window->picture->rscreen->h : 200;
                w2 *= arguments.zoom;
                h2 *= arguments.zoom;
            } else {
                int scale = arguments.zoom;
                w2 = main_window->picture->rscreen ? main_window->picture->rscreen->w : 320;
                h2 = main_window->picture->rscreen ? main_window->picture->rscreen->h : 200;
                if (!scale) {
                    scale = 1;
                    int sw = Fl::w() * 0.618, sh = Fl::h() * 0.618 - FL_NORMAL_SIZE * 2;
                    while (w2 * (scale + 1) < sw && h2 * (scale + 1) < sh && scale < 13) {
                        scale++;
                    }
                }
                w2 *= scale;
                h2 *= scale;
            }
            h2 += main_window->menu->h();

            x += (w - w2) / 2;
            y += (h - h2) / 2;

            main_window->resize(x, y, w2, h2);
        }
    }

    int handle(int e) {
        switch (e) {
        case FL_KEYDOWN:
            if (!Fl::event_shift() && !Fl::event_ctrl() && !Fl::event_alt() && !Fl::event_command()) {
                switch (Fl::event_key()) {
                case '=': zoom2_cb(this, (void *)3); return 1;
                case FL_Enter: quit_cb(this, NULL); return 1;
                default: break;
                }
            }
            if (!Fl::event_shift() && !Fl::event_alt() && Fl::event_command()) {
                switch (Fl::event_key()) {
                case 's': save_cb(this, NULL); return 1;
                case 'o': open_cb(this, NULL); return 1;
                default: break;
                }
            }
            if (Fl::event_shift() && !Fl::event_ctrl() && !Fl::event_alt() && !Fl::event_command()) {
                switch (Fl::event_key()) {
                case '9': case '8': case '7': case '6': case '5': case '4': case '3': case '2': case '1':
                case '0': zoom_cb(this, (void *)(Fl::event_key() - '0')); resize_cb(this, NULL); return 1;
                default: break;
                }
            }
            break;
        }
        return Fl_Window::handle(e);
    }
};

extern "C" int main2(int, char *[]);

int main2(int argc, char *argv[])
{
    char cfgdir[MAX_PATH];
    float dx, dy;
    Fl::visual(FL_RGB8);
    Fl::screen_dpi(dx, dy);
    if (dy > 600) dy = 600;
    if (dy < 50) dy = 50;
    NORMAL_SIZE = FL_NORMAL_SIZE;
    MY_SIZE = FL_NORMAL_SIZE = (int)(FL_NORMAL_SIZE * dy / 96);

    fl_message_title_default("View64 " VERSION);

    config_reset(&arguments);
    get_user_config_file(cfgdir, sizeof(cfgdir), "view64");
    if (cfgdir[0]) {
        FILE *f1 = fl_fopen(cfgdir, "r");
        if (f1) {
            ini_parse_file(f1, ini_handler, &arguments);
            fclose(f1);
        }
    }
    configure(argc, argv, &arguments);
    arguments.zoom = arguments.zoom > 13 ? 13 : arguments.zoom;

    main_window = new Main_window();
    Fl::run();
    delete main_window;

    if (config_changed(&arguments)) {
        if (cfgdir[0]) {
            FILE *f1 = fl_fopen(cfgdir, "wt");
            if (f1) {
                ini_save_file(f1, &arguments);
                fclose(f1);
            }
        }
    }
    free_config(&arguments);
    return 0;
}
