/*
    View64 C64 image viewer
    (c) 1995-2013 Soci/Singular (soci at c64.rulez.org)

    $Id: view64.c 46 2015-05-11 17:50:36Z soci $

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License along
    with this program; if not, write to the Free Software Foundation, Inc.,
    51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

*/

#include <stdio.h>

#include "SDL.h"
#include "libview64.h"
#include "tvrender.h"
#include "view64_config.h"
#include "main.h"
#include "cfgpath.h"
#include "ini.h"

static SDL_Surface *create_surface(view64_screen *rscreen, unsigned int vmode, int scale, const rgb_color *palette) {
    SDL_Surface *pic;
    tvrender_image out;
    pic = SDL_CreateRGBSurface(0, rscreen->w * scale, rscreen->h * scale, 24, 0xff, 0xff << 8, 0xff << 16, 0);
    if (pic == NULL) {
        SDL_LogError(SDL_LOG_CATEGORY_APPLICATION, "Couldn't create RGB surface: %s\n", SDL_GetError());
        return NULL;
    }
    out.d = (unsigned char *)pic->pixels;
    out.h = pic->h;
    out.w = pic->w;
    out.pitch = pic->pitch;
    create_tvpicture(rscreen, vmode, scale, palette, &out);
    return pic;
}

static void SetWindowPosition(SDL_Window *win, const SDL_Rect *windim) {
    int x = windim->x - windim->w / 2;
    int y = windim->y - windim->h / 2;
    if (x < 0) x = 0;
    if (y < 0) y = 0;
    SDL_SetWindowSize(win, windim->w, windim->h);
    SDL_SetWindowPosition(win, x, y);
}

static void render_texture(SDL_Renderer *renderer, SDL_Window *win, view64_screen *rscreen, unsigned int vmode, const rgb_color *palette, const SDL_Rect *windim, int ab) {
    static SDL_Texture *texture = NULL;
    static int texture_scale = 0;
    static unsigned int texture_video = VIDEO_RGB;
    static SDL_Surface *pic2 = NULL;
    int h, w;
    int scale, scale2;
    SDL_Rect dst;

    if (win == NULL) {
        if (pic2) SDL_FreeSurface(pic2);
        if (texture) SDL_DestroyTexture(texture);
        return;
    }
    if (ab) {
        w = windim->w; h = windim->h;
    } else {
        SDL_GetWindowSize(win, &w, &h);
    }

    scale = (w - 1) / rscreen->w;
    scale2 = (h - 1) / rscreen->h;
    if (scale2 < scale) scale = scale2;
    scale++;
    if (scale < 1) scale = 1;

    h = rscreen->h * scale;
    w = rscreen->w * scale;
    SDL_RenderSetLogicalSize(renderer, w, h);
    SDL_RenderClear(renderer);

    if (texture_scale != scale || texture_video != vmode) {
        if (pic2) SDL_FreeSurface(pic2);
        pic2 = create_surface(rscreen, vmode, scale, palette);
        if (pic2 == NULL) return;
        if (texture) SDL_DestroyTexture(texture);
        texture = SDL_CreateTextureFromSurface(renderer, pic2);
        texture_scale = scale;
        texture_video = vmode;
    }
    dst.x = 0;
    dst.y = 0;
    dst.w = w;
    dst.h = h;
    if (ab) {
        SetWindowPosition(win, windim);
    }
    SDL_RenderCopy(renderer, texture, NULL, &dst);
    SDL_RenderPresent(renderer);
}

extern int main2(int, char *[]);

int main2(int argc, char *argv[])
{
    char cfgdir[MAX_PATH];
    SDL_Window *win;
    SDL_Renderer *renderer;
    SDL_DisplayMode mode;
    SDL_Event event;
    unsigned char *mem;
    view64_screen *rscreen;
    int done;
    FILE *f1;
    const char *filename = NULL;
    size_t size;
    int vi, scale, scale2;
    view64_format format;
    char *window_title;
    const char *format_name;
    int width, height, zoom;
    rgb_color *palette;
    Uint32 winflags;
    unsigned int vmode;
    arguments_t arguments;
    Uint32 check_resize;
    int ab;
    SDL_Rect windim;

    config_reset(&arguments);
    get_user_config_file(cfgdir, sizeof(cfgdir), "view64");
    if (cfgdir[0]) {
        FILE *f1 = fopen2(cfgdir, "r");
        if (f1) {
            ini_parse_file(f1, ini_handler, &arguments);
            fclose(f1);
        }
    }
    configure(argc, argv, &arguments);
    filename = arguments.filename;
    width = arguments.width > 4160 ? 4160 : arguments.width;
    height = arguments.height > 2600 ? 2600 : arguments.height;
    winflags = SDL_WINDOW_RESIZABLE;
    vmode = arguments.vmode;
    zoom = arguments.zoom > 13 ? 13 : arguments.zoom;

    if (!filename) {
        puts("Usage: view64 [OPTIONS...] PICTURE\n"
             "Try `view64 --help' or `view64 --usage' for more information.\n");
        return 1;
    }

    if ((f1 = fopen2(filename, "rb")) == NULL) {
        fprintf(stderr, "Couldn't open file \"%s\"\n", filename);
        free_config(&arguments);
        return 1;
    }

    if (fseek(f1, 0, SEEK_END)) {
        size = 65536;
    } else {
        size = ftell(f1);
        if (size < 65536) size = 65536;
    }
    rewind(f1);
    mem = (unsigned char *)calloc(1, size);
    if (mem == NULL) {
        fclose(f1);
        return 1;
    }

    size = fread(mem, 1, size, f1);
    fclose(f1);

    format = view64_format_detect(filename, mem, size);
    if (format == VIEW64_UNKNOWN) {
        fprintf(stderr, "%s: Image format not recognized\n", filename);
        free(mem);
        free_config(&arguments);
        return 1;
    }
    rscreen = view64_decode(format, mem, size, 0);
    free(mem);
    if (rscreen == NULL) {
        fprintf(stderr, "%s: Can't decode image\n", filename);
        free_config(&arguments);
        return 1;
    }

    if (width < 1 || zoom > 0) {
        if (height < 1 || zoom > 0) {
            /* later */
        } else {
            width = rscreen->w * height / rscreen->h;
        }
    } else if (height < 1) {
        height = (double)rscreen->h * width / rscreen->w;
    } else {
        /* all set */
    }

    if (SDL_Init(arguments.save ? 0 : SDL_INIT_VIDEO) < 0) {
        SDL_LogCritical(SDL_LOG_CATEGORY_APPLICATION, "Couldn't initialize SDL: %s\n", SDL_GetError());
        free_config(&arguments);
        return 1;
    }

    f1 = arguments.palette_filename && arguments.palette_filename[0] ? fopen2(arguments.palette_filename, "rb") : NULL;
    palette = load_palette(f1);
    if (f1) fclose(f1);

    if (arguments.save) {
        char *newname = (char *)malloc(strlen(filename) + 5);
        if (newname != NULL) {
            if (zoom < 1) {
                zoom = 1;
                while (rscreen->w * (zoom + 1) <= width && rscreen->h * (zoom + 1) <= height) zoom++;
            }
            SDL_Surface *surface = create_surface(rscreen, vmode, zoom, palette);
            if (surface) {
                strcpy(newname, filename);
                strcat(newname, ".bmp");
                SDL_SaveBMP(surface, newname);
                SDL_FreeSurface(surface);
            }
            free(newname);
        }
        view64_free_screen(rscreen);
        free(palette);
        free_config(&arguments);
        SDL_Quit();
        return 0;
    }
    format_name = view64_format_name(format);
    window_title = (char *)malloc(strlen(format_name) + strlen("View64 " VERSION " - ") + 4 + strlen(filename));
    strcpy(window_title, "View64 " VERSION " - ");
    strcat(window_title, format_name);
    strcat(window_title, " - ");
    strcat(window_title, filename);

    if ((win = SDL_CreateWindow(window_title, SDL_WINDOWPOS_UNDEFINED, SDL_WINDOWPOS_UNDEFINED, width, height, winflags | SDL_WINDOW_HIDDEN)) == NULL) {
        SDL_LogCritical(SDL_LOG_CATEGORY_APPLICATION, "Couldn't create window: %s\n", SDL_GetError());
        free(palette);
        free_config(&arguments);
        SDL_Quit();
        return 1;
    }
    free(window_title);
    renderer = SDL_CreateRenderer(win, -1, SDL_RENDERER_PRESENTVSYNC);
    SDL_SetHint(SDL_HINT_RENDER_SCALE_QUALITY, "best");

    vi = SDL_GetWindowDisplayIndex(win);
    SDL_GetCurrentDisplayMode(vi, &mode);

    if (zoom > 0) {
        scale = zoom;
        width = rscreen->w;
        height = rscreen->h;
    } else {
        scale = 1;
        if (width < 1 && height < 1) {
            width = rscreen->w;
            height = rscreen->h;
            while (width * (scale + 1) < mode.w * 0.618 && height * (scale + 1) < mode.h * 0.618) {
                scale++;
            }
        }
    }

    windim.w = width * scale; windim.h = height * scale;
    if (arguments.x < 0 || arguments.y < 0) {
        arguments.x = mode.w / 2; arguments.y = mode.h / 2;
    }
    windim.x = arguments.x; windim.y = arguments.y;
    if (windim.x - windim.w / 2 > mode.w) windim.x = mode.w + windim.w / 2;
    if (windim.y - windim.h / 2 > mode.h) windim.y = mode.h + windim.h / 2;
    if (arguments.fullscreen) {
        SDL_SetWindowSize(win, windim.w, windim.h);
        SDL_SetWindowSize(win, 1, 1);
        winflags |= SDL_WINDOW_FULLSCREEN_DESKTOP;
        SDL_SetWindowBordered(win, SDL_FALSE);
        SDL_ShowCursor(SDL_FALSE);
        SDL_SetWindowFullscreen(win, winflags);
    } else {
        SetWindowPosition(win, &windim);
    }
    SDL_ShowWindow(win);

    ab = !!zoom;
    check_resize = SDL_GetTicks() + 500;
    done = 0;
    while (!done)
    {
        SDL_WaitEvent(&event);

        switch (event.type) {
        case SDL_KEYDOWN:
            switch (event.key.keysym.sym) {
            case SDLK_ESCAPE:
                done = 1;
                break;
            case SDLK_p:
                vmode = (vmode & ~VIDEO_MASK) | VIDEO_PAL;
                goto resize2;
            case SDLK_n:
                vmode = (vmode & ~VIDEO_MASK) | VIDEO_NTSC;
                goto resize2;
            case SDLK_r:
                vmode = (vmode & ~VIDEO_MASK) | VIDEO_RGB;
                goto resize2;
            case SDLK_t:
                vmode ^= VIDEO_SHADOWMASK;
                goto resize2;
            case SDLK_l:
                vmode ^= VIDEO_SCANLINES;
                goto resize2;
            case SDLK_c:
                vmode ^= VIDEO_CHROMA_NOISE;
                goto resize2;
            case SDLK_i:
                vmode ^= VIDEO_NOLACE;
                goto resize2;
            case SDLK_m:
                vmode ^= VIDEO_MONOCHROME;
                goto resize2;
            case SDLK_9: zoom = scale = 9; goto resize;
            case SDLK_8: zoom = scale = 8; goto resize;
            case SDLK_7: zoom = scale = 7; goto resize;
            case SDLK_6: zoom = scale = 6; goto resize;
            case SDLK_5: zoom = scale = 5; goto resize;
            case SDLK_4: zoom = scale = 4; goto resize;
            case SDLK_3: zoom = scale = 3; goto resize;
            case SDLK_2: zoom = scale = 2; goto resize;
            case SDLK_1: zoom = scale = 1; goto resize;
            case SDLK_0: width = rscreen->w; 
                         height = rscreen->h; 
                         scale = 1;
                         while (width * (scale + 1) < mode.w * 0.618 && height * (scale + 1) < mode.h * 0.618) {
                             scale++;
                         }
                         arguments.width = arguments.height = zoom = 0; 
                         check_resize = SDL_GetTicks() + 500;
                         windim.w = width * scale; windim.h = height * scale;
                         arguments.x = windim.x = mode.w / 2; arguments.y = windim.y = mode.h / 2;
                         ab = 1;
                         goto resize2;
                {
                resize:
                    ab = 1;
                    arguments.width = arguments.height = 0;
                    check_resize = SDL_GetTicks() + 500;
                    width = rscreen->w * scale;
                    height = rscreen->h * scale;
                resize3:
                    scale = (width - 1) / rscreen->w;
                    scale2 = (height - 1) / rscreen->h;
                    if (scale2 < scale) scale = scale2;
                    scale++;
                    if (scale < 1) scale = 1;
                    windim.w = width; windim.h = height;
                resize2:
                    if (winflags & SDL_WINDOW_FULLSCREEN_DESKTOP) {
                        SDL_Rect windim2 = {0, 0, mode.w, mode.h};
                        render_texture(renderer, win, rscreen, vmode, palette, &windim2, 1);
                    } else {
                        render_texture(renderer, win, rscreen, vmode, palette, &windim, ab);
                    }
                }
                break;
            case SDLK_f:
                winflags ^= SDL_WINDOW_FULLSCREEN_DESKTOP;
                if (winflags & SDL_WINDOW_FULLSCREEN_DESKTOP) {
                    SDL_SetWindowBordered(win, SDL_FALSE);
                    SDL_SetWindowSize(win, 1, 1);
                    SDL_ShowCursor(SDL_FALSE);
                } 
                SDL_SetWindowFullscreen(win, winflags);
                if (!(winflags & SDL_WINDOW_FULLSCREEN_DESKTOP)) {
                    SDL_SetWindowBordered(win, SDL_TRUE);
                    SetWindowPosition(win, &windim);
                    SDL_ShowCursor(SDL_TRUE);
                }
                check_resize = SDL_GetTicks() + 500;
                goto resize2;
            case SDLK_s:
                {
                    char *newname = (char *)malloc(strlen(filename) + 5);
                    if (newname != NULL) {
                        SDL_Surface *surface = create_surface(rscreen, vmode, scale, palette);
                        if (surface) {
                            strcpy(newname, filename);
                            strcat(newname, ".bmp");
                            SDL_SaveBMP(surface, newname);
                            SDL_FreeSurface(surface);
                        }
                        free(newname);
                    }
                }
                break;
            case SDLK_UP:
                windim.y -= 10;
                if (windim.y < 0) windim.y = 0;
                arguments.x = windim.x; arguments.y = windim.y;
                goto resize2;
            case SDLK_LEFT:
                windim.x -= 10;
                if (windim.x < 0) windim.x = 0;
                arguments.x = windim.x; arguments.y = windim.y;
                goto resize2;
            case SDLK_DOWN:
                windim.y += 10;
                if (windim.y > mode.h) windim.y = mode.h;
                arguments.x = windim.x; arguments.y = windim.y;
                goto resize2;
            case SDLK_RIGHT:
                windim.x += 10;
                if (windim.x > mode.w) windim.x = mode.w;
                arguments.x = windim.x; arguments.y = windim.y;
                goto resize2;
            default: /*fprintf(stderr, SDL_GetKeyName(event.key.keysym.sym));*/break;
            }
            break;
        case SDL_WINDOWEVENT:
            switch (event.window.event) {
            case SDL_WINDOWEVENT_EXPOSED: goto resize2;
            case SDL_WINDOWEVENT_RESIZED:
                if (SDL_TICKS_PASSED(SDL_GetTicks(), check_resize)) {
                    windim.x -= windim.w / 2;
                    windim.y -= windim.h / 2;
                    width = arguments.width = event.window.data1;
                    height = arguments.height = event.window.data2;
                    windim.x += width / 2;
                    windim.y += height / 2;
                    zoom = 0;
                    render_texture(renderer, win, rscreen, vmode, palette, &windim, 0);
                    ab = 0;
                    goto resize3;
                }
                break;
            case SDL_WINDOWEVENT_MOVED:
                if (SDL_TICKS_PASSED(SDL_GetTicks(), check_resize)) {
                    windim.x = arguments.x = event.window.data1 + windim.w / 2;
                    windim.y = arguments.y = event.window.data2 + windim.h / 2;
                }
                break;
            }
            break;
        case SDL_QUIT:
            done = 1;
            break;
        default:
            break;
        }
    }
    arguments.fullscreen = !!(winflags & SDL_WINDOW_FULLSCREEN_DESKTOP);
    arguments.vmode = vmode;
    arguments.zoom = zoom;

    render_texture(NULL, NULL, NULL, 0, NULL, NULL, 0);
    view64_free_screen(rscreen);
    free(palette);
    SDL_DestroyWindow(win);

    if (config_changed(&arguments)) {
        get_user_config_file(cfgdir, sizeof(cfgdir), "view64");
        if (cfgdir[0]) {
            FILE *f1 = fopen2(cfgdir, "wt");
            if (f1 == NULL) {
                fprintf(stderr, "Couldn't save file \"%s\"\n", cfgdir);
            } else {
                ini_save_file(f1, &arguments);
                fclose(f1);
            }
        }
    }
    free_config(&arguments);
    SDL_Quit();
    return 0;
}
